<?php
include_once "ProxmoxClient.php";
include_once "common/common.class.php";

/**
    @file
    @brief Fichier drivers pour les clusters Proxmox
    @details
    Ce fichier contient un ensemble de classes qui étendent les classes du
    fichier common.class.php pour l'hyperviseur Proxmox.

    Ce fichier inclut le fichier ProxmoxClient qui gére les requêtes envoyés à
    un hyperviseur Proxmox.
    Les classes dans ce fichier manipulent la classe ProxmoxClient de ce fichier
    et formattent ces données en respectant les conventions décrient dans common.class.php
    @author Julien George
*/

class ProxmoxCluster extends Cluster{
    protected $driver;

    protected $cluster_status=NULL;
    protected $cluster_raw = NULL;
    protected $object_cache = array();

    protected $resource_data = array();

    public function __construct($settings){
        $this->driver = new ProxmoxClient($settings,TRUE);
    }

    public function cleanCache(){
            $this->cluster_status = NULL;
            $this->cluster_raw = NULL;
            $this->object_cache = array();
            $this->resource_data = array();
    }

    public function getRawData($key,$url,$params=NULL){
        if( !isset($this->resource_data[$key]) ){
            $this->resource_data[$key] = $this->driver->get($url,$params);
        }
        return $this->resource_data[$key];
    }

    public function jsonSerialize(){
        if( $this->cluster_status == NULL ){
            $this->cluster_status = array_fill_keys(array('name','master','is_cluster','is_ha'),NULL);

            $cluster = $this->getRawData('cluster_status','cluster/status');

            $this->cluster_raw =array();
            foreach($cluster as $row){
                $this->cluster_raw[$row['id']] = $row;
            }

            if (count ($cluster) == 1)
              {
                      $this->cluster_status['name'] = NULL;
                      $this->cluster_status['master'] = $row['id'];
                      $this->cluster_status['is_cluster'] = FALSE;
                      $this->cluster_status['is_ha'] = FALSE;
              }
              else
              {
                      $ha = $this->getRawData('cluster_ha','cluster/ha/status/current');
                      foreach($cluster as $rcluster){
                              if($rcluster['type'] == 'cluster'){
                                      $this->cluster_status['name'] = $rcluster['name'];
                                      $this->cluster_status['is_cluster'] = TRUE;
                              }
                              foreach($ha as $rha){
                                      if( $rha['type'] == 'quorum'){
                                              $this->cluster_status['is_ha'] = ( $rha['status'] == 'OK' );
                                      }
                                      else if( $rha['type'] == 'master' ){
                                              $this->cluster_status['master'] = 'node/'.$rha['node'];
                                      }
                              }
                              //Cas rare/particulier :
                              //Si il n'y a pas de master renseigné, alors il n'y a probablement jamais eu de HA dans ce cluster (aucune VM dans ce mode)
                              if( $this->cluster_status['master']  == NULL ){
                                  $this->cluster_status['is_ha'] = FALSE;
                              }
                      }
              }
        }
        return $this->cluster_status;
    }

    public function getDriver(){
        return $this->driver;
    }

    public function getName(){
        return $this->jsonSerialize()['name'];
    }
    public function getMasterHost(){
        return$this->jsonSerialize()['master'];
    }

    public function isCluster(){
        return $this->jsonSerialize()['is_cluster'];
    }

    public function isHA(){
        return $this->jsonSerialize()['is_ha'];
    }

    public function getHosts():array{
        if( !isset($this->object_cache['host'])){
            $host_list = array();
            foreach($this->getRawData('nodes','nodes') as $raw_row){
                $host_list[] = new ProxmoxHost($raw_row['id'],$this);
            }
            $this->object_cache['host'] = $host_list;
        }
        return $this->object_cache['host'];
    }

    public function getHostByName($name){
        foreach($this->getHosts() as $host){
            if( $host->getName() == $name)
            {
                return $host;
            }
        }
        return NULL;
    }

    public function getVMs():array{
        if( !isset($this->object_cache['vm'])){
            $vm_list = array();
            foreach($this->getHosts() as $host){
                $vm_list = array_merge($vm_list,$host->getVMs());
            }
            $this->object_cache['vm']=$vm_list;
        }
        return $this->object_cache['vm'];
    }

    public function getTemplates():array{
        if( !isset($this->object_cache['templates'])){
            $vm_list = array();
            foreach($this->getHosts() as $host){
                $vm_list = array_merge($vm_list,$host->getTemplates());
            }
            $this->object_cache['templates']=$vm_list;
        }
        return $this->object_cache['templates'];
    }
    public function getStorages($shared=NULL):array{
        if( !isset($this->object_cache['storage'])){
            $storage_list = array();
            $resources = $this->getRawData('cluster_resources_storage','cluster/resources',array('type'=>'storage'));
            foreach($resources as $raw_row){
                if( $raw_row['shared'] == 1 )   $id = 'cluster/'.$raw_row['storage'];
                else                            $id = $raw_row['node'].'/'.$raw_row['storage'];

                $storage_list[$id] = new ProxmoxStorage($id,$this);
            }
            $this->object_cache['storage'] = array_values($storage_list);
        }
        if( $shared === NULL || !is_bool($shared) ){
            return $this->object_cache['storage'];
        }
        else{
            $ret = array();
            foreach($this->object_cache['storage'] as $storage){
                if( $storage->isShared() == $shared ){
                    $ret[] = $storage;
                }
            }
            return $ret;
        }
    }

    private function getHostNameByVMID($vmid){
        foreach($this->getHosts() as $host){
            $vm_list = array_merge($host->getVMData(FALSE),$host->getVMData(TRUE));
            foreach($vm_list as $vm){
                if( $vm['vmid'] == $vmid ){
                    return $host->getName();
                }
            }
        }
        return NULL;
    }

    public function getVMFromSnapshot($snapid){
        return $this->getVM(explode('/',$snapid)[0]);
    }

    function getVMByName($name){
        $resources = $this->getRawData('cluster_resources_vm','cluster/resources',array('type'=>'vm'));
        $ret = array();
        foreach($resources as $row) {
            if( $row['name'] == $name && $row['template'] != 1 ){
                $host = new ProxmoxHost('node/'.$row['node'],$this);
                $ret[] = new ProxmoxVM($row,$host);
            }
        }
        return $ret;
    }

    public function getVM($vmid,$refresh=FALSE){
        if($refresh){
            //On efface les caches
            unset($this->object_cache['host']); //Les hosts car ils contiennent les données VMs
            unset($this->object_cache['vm']); //Les VMs elles mêmes
            unset($this->resource_data['cluster_resources_vm']); //Les données brutes des VMs qui sont stockées ici
        }
        $vm_list = $this->getVMs();
        foreach($vm_list as $vm){
            if( $vm->getId() == $vmid ){
                return $vm;
            }
        }
        return NULL;
    }

    public function getNextVMID(){
        $vmid=0;
        $vmid_list = array();
        foreach($this->getHosts() as $host){
            $vm_list = array_merge($host->getVMData(FALSE),$host->getVMData(TRUE));
            foreach($vm_list as $vm){
                $vmid_list[$vm['vmid']]= TRUE;
            }
        }
        ksort($vmid_list);
        $i=100;
        while(isset($vmid_list[$i])){
            $i++;
        }
        return $i;
    }

    public function cloneVM($source,$target_name,$full=TRUE){
        //Analyse de la source
        $vmid=NULL;
        $host=NULL;
        if( is_string($source)){
            $vmid = $source;
            $host = $this->getHostNameByVMID($vmid);
        }
        elseif(is_array($source)&& isset($source['id'])){
            $vmid=$source['id'];
            if(isset($source['host'])){
                $host = $source['host'];
            }
            else
            {
                $host = $this->getHostNameByVMID($vmid);
            }
        }
        elseif(is_object($source)){
            $vmid = $source->getId();
            $host = $source->getHost()->getName();
        }

        if( $vmid == NULL || $host == NULL ){
            throw new Exception("Unable to found the VM/template source\n");
        }

        //Start
        $new_vmid = $this->getNextVMID();
        try{
            $this->driver->post("/nodes/$host/qemu/$vmid/clone",array('newid'=>$new_vmid,'name'=>$target_name,'full'=>$full));
        }
        catch(Exception $ex){
            if( $full == FALSE && strpos($ex->getMessage(), 'Http error : 500') !== FALSE ){
                throw new Exception('Unable to create linked clone. Source must be template');
            }
            else throw $ex;
        }
        //On efface les caches
        unset($this->object_cache['host']); //Les hosts car ils contiennent les données VMs
        unset($this->object_cache['vm']); //Les VMs elles mêmes
        unset($this->resource_data['cluster_resources_vm']); //Les données brutes des VMs qui sont stockées ici
    }
}

class ProxmoxHost extends Host{
    protected $cluster;

    protected $resource_data=NULL;
    protected $rrd_data=NULL;

    protected $vm_data = NULL;
    protected $template_data = NULL;

    public function __construct($id,ProxmoxCluster $cluster){
        $this->cluster = $cluster;
        $this->setId($id);
        $this->setName(explode('/',$id)[1]);
        //$this->setDescription('');
        $this->setType('hypervisor');
        $this->setManufacturer('Proxmox');
        $this->setOS('Proxmox');
        //$this->setOSVersion('');
    }

    private function getResourceData(){
        $host_id = $this->getId();
        if( $this->resource_data == NULL ){
            $resources = $this->cluster->getRawData('nodes','nodes');
            foreach($resources as $res){
                if( $res['id'] == $host_id ){
                    $this->resource_data = $res;
                    break;
                }
            }
        }
        if(empty($this->resource_data)) throw new Exception("Not found Host resource $host_id");
        return $this->resource_data;
    }

    public function getRRDValues($key,$url,$nb_min){
        $rrd =$this->cluster->getRawData($key,$url,array('timeframe'=>'hour'));

        $tmp = array();
        foreach($rrd as $row){
                if( count($row) <= 1 ) continue;
                $tmp[$row['time']]=$row;
        }
        ksort($tmp);
        $tmp=array_slice($tmp, -$nb_min,NULL,TRUE); //On garde les x dernieres minutes

        $max = array_fill_keys(array_keys(current($tmp)),0);
        unset($max['time']);
        foreach($tmp as $row){
                foreach($max as $key => $value){
                        if( !isset($max[$key])) return NULL;
                        if( $max[$key] < $row[$key]){
                                $max[$key] = round($row[$key]);
                        }
                }
        }
        return $max;
    }

    private function getRRDData(){
        if( $this->rrd_data == NULL ){
            $nodename = $this->getName();
            $this->rrd_data = $this->getRRDValues($nodename.'_rrd',"nodes/$nodename/rrddata",2);
        }
        return $this->rrd_data;
    }

    public function getVMData($readonly=FALSE){
        if($this->vm_data == NULL ){
            $vm_list = array();
            $resources = $this->cluster->getRawData('cluster_resources_vm','cluster/resources',array('type'=>'vm'));
            $this->vm_data = array();
            $this->template_data = array();
            foreach($resources as $raw_row){
                if( $raw_row['node'] == $this->getName() ){
                    if( !isset($raw_row['template']) && isset($raw_row['lock'])){
                        if( !isset($raw_row['name'])){
                            $raw_row['name'] = 'VM '.$raw_row['vmid'];
                        }
                        $this->vm_data[] = $raw_row;
                    }
                    elseif( $raw_row['template'] != 1 ){
                        $this->vm_data[] = $raw_row;
                    }
                    else{
                        $this->template_data[] = $raw_row;
                    }
                }
            }
        }
        if( !$readonly ) return $this->vm_data;
        else return $this->template_data;
    }


    public function getCluster(){
        return $this->cluster;
    }

    public function getAddress(){
        if( parent::getAddress() == NULL){
            $raw_data = $this->cluster->getRawData('cluster_status','cluster/status');
            foreach($raw_data as $row) {
                if( $row['type'] == 'node' && $row['name'] == $this->getName()){
                    if(isset($row['ip'])){
                        $this->setAddress($row['ip']);
                    }
                    break;
                }
            }
        }
        return parent::getAddress();
    }

    public function getUpTime():int{
        return $this->getResourceData()['uptime'];
    }

    public function getCpusTotal():int{
        return $this->getResourceData()['maxcpu'];
    }

    public function getCpusUsage():float{
        return $this->getResourceData()['cpu'];
    }

    public function getMemTotal():int{
        return $this->getResourceData()['maxmem'];
    }

    public function getMemUsage():int{
        return $this->getResourceData()['mem'];
    }

    public function getNetInUsage():int{
        return $this->getRRDData()['netin'];
    }

    public function getNetOutUsage():int{
        return $this->getRRDData()['netout'];
    }

    public function getNetInterfaces()
    {
      $pif_speed = array();
      $pif_speed['global']['rx'] = $this->getNetInUsage();
      $pif_speed['global']['tx'] = $this->getNetOutUsage();

      $nodename = $this->getName();
      $raw_data = $this->cluster->getRawData($nodename.'_net',"/nodes/$nodename/network");

      foreach($raw_data as $row)
      {
          $dev_name = $row['iface'];
          $retRow['device'] = $dev_name;
          if(isset($row['address']))
          {
             $retRow['ip'] = $row['address'];
          }
          else
          {
              $retRow['ip'] = NULL;
          }
          $retRow['mac'] = NULL;
          if(isset($row['mtu']))
          {
             $retRow['mtu'] = $row['mtu'];
          }
          else
          {
              $retRow['mtu'] = NULL;
          }
          $retRow['rx'] = NULL;
          $retRow['tx'] = NULL;
          $retRow['speed'] = NULL;
          $pif_speed[$dev_name]=$retRow;
      }
      return $pif_speed;
    }

    public function isAlive():bool{
        return $this->getResourceData()['status'] == 'online';
    }

    public function getVMs():array{
        $vm_list = array();
        foreach($this->getVMData(FALSE) as $raw_row){
            $vm_list[] = new ProxmoxVM($raw_row,$this);
        }
        return $vm_list;
    }

    public function getTemplates(){
        $vm_list = array();
        foreach($this->getVMData(TRUE) as $raw_row){
            $tpl = new ProxmoxVM($raw_row,$this);
            $vm_list[] = $tpl->toTemplateArray();
        }
        return $vm_list;
    }
}

class ProxmoxVM extends VM{
    protected $cluster;
    protected $host;

    protected $resource_data=NULL;
    protected $on_host_data = NULL;
    protected $config_data = NULL;
    protected $rrd_data=NULL;
    protected $snap_data=NULL;

    private static $enum_os_type = array(
        'unknown'  => 'unknown',
        'other'    => 'unspecified',
        'wxp'      => 'Microsoft Windows',
        'w2k'      => 'Microsoft Windows',
        'w2k3'     => 'Microsoft Windows',
        'w3k8'     => 'Microsoft Windows',
        'wvista'   => 'Microsoft Windows',
        'win7'     => 'Microsoft Windows',
        'win8'     => 'Microsoft Windows',
        'win10'    => 'Microsoft Windows',
        'l24'      => 'Unix/Linux',
        'l26'      => 'Unix/Linux',
        'solaris'  => 'Unix/Linux',
    );

    private static $enum_os_version = array(
        'unknown'  => 'unknown',
        'other'    => 'unspecified',
        'wxp'      => 'XP',
        'w2k'      => '2000',
        'w2k3'     => '2003',
        'w3k8'     => '2008',
        'wvista'   => 'Vista',
        'win7'     => '7',
        'win8'     => '8/2012/2012r2',
        'win10'    => '10/2016',
        'l24'      => '2.4 Kernel',
        'l26'      => '2.6 - 5.X Kernel',
        'solaris'  => 'Solaris kernel',
    );

    private static $enum_power_states = array(
        'started' => 'running',
        'running' => 'running',
        'stopped' => 'stopped',
        'halted'  => 'stopped',
        'paused'  => 'paused',
    );

    public function __construct($raw_data,ProxmoxHost $host){
        $this->resource_data = $raw_data;
        $this->cluster = $host->getCluster();
        $this->host = $host;

        $this->setId($raw_data['vmid']);
        $this->setName($raw_data['name']);
        $this->setType('VM');
        $this->setManufacturer('Proxmox');
    }

    public function getVMRawConfig(){
        if( $this->config_data == NULL ){
            $node = $this->getHost()->getName();
            $id = $this->getId();
            $this->config_data = $this->cluster->getDriver()->get("/nodes/$node/qemu/$id/config");
        }
        return $this->config_data;
    }

    private function getRRDData($key=NULL){
        if( $this->rrd_data == NULL ){
            $nodename = $this->getHost()->getName();
            $id = $this->getId();
            $this->rrd_data = $this->host->getRRDValues($id.'_rrd',"/nodes/$nodename/qemu/$id/rrddata",2);
        }
        if( $key == NULL){
            return $this->rrd_data;
        }
        else if( !isset($this->rrd_data[$key]) ){
            return NULL;
        }
        else{
            return $this->rrd_data[$key];
        }
    }

    private function getVMSnapData(){
        if( $this->snap_data == NULL ){
            $node = $this->getHost()->getName();
            $id = $this->getId();
            $this->snap_data = $this->cluster->getDriver()->get("/nodes/$node/qemu/$id/snapshot");
        }
        return $this->snap_data;
    }

    public function getHost(){
        return $this->host;
    }

    public function getOS(){
        if( !isset($this->getVMRawConfig()['ostype'])){
            $key = 'unknown';
        }
        else{
            $key = $this->getVMRawConfig()['ostype'];
        }
        return self::$enum_os_type[$key];
    }

    public function getOSVersion(){
        if( !isset($this->getVMRawConfig()['ostype'])){
            $key = 'unknown';
        }
        else{
            $key = $this->getVMRawConfig()['ostype'];
        }
        return self::$enum_os_version[$key];
    }

    public function getUpTime(){
        if( !isset($this->resource_data['uptime'])) return NULL;
        return $this->resource_data['uptime'];
    }

    public function getCpuPlatform(){
        if( !isset($this->getVMRawConfig()['cpu'])){
            return NULL;
        }
        else{
            return $this->getVMRawConfig()['cpu'];
        }
    }

    public function isAgent(){
        $raw_data = $this->getVMRawConfig();
        return (isset($raw_data['agent']) && explode(',',$raw_data['agent'])[0] == 1);
    }

    public function getCpusTotal(){
        if( !isset($this->resource_data['maxcpu'])) return NULL;
        return $this->resource_data['maxcpu'];
    }

    public function getCpusUsage(){
        if( !isset($this->resource_data['cpu'])) return NULL;
        return $this->resource_data['cpu'];
    }

    public function getMemTotal(){
        if( !isset($this->resource_data['maxmem'])) return NULL;
        return $this->resource_data['maxmem'];
    }

    public function getMemUsage(){
        if( !isset($this->resource_data['mem'])) return NULL;
        return $this->resource_data['mem'];
    }

    public function getNetInUsage(){
        return $this->getRRDData('netin');
    }

    public function getNetOutUsage(){
        return $this->getRRDData('netout');
    }

    public function getState(){
        if( !isset(self::$enum_power_states[$this->resource_data['status']])) return NULL;
        return self::$enum_power_states[$this->resource_data['status']];
    }

    public function isLocked($refresh=FALSE){
        $status = $this->getLockStatus($refresh);
        return ( $status != NULL && $status != 'suspended');
    }

    public function getLockStatus($refresh=FALSE){
        if( $refresh ){
            $this->resource_data['lock'] = $this->cluster->getVM($this->getID(),TRUE)->getLockStatus();
        }
        if( empty($this->resource_data['lock'])) return NULL;
        return $this->resource_data['lock'];
    }

    public function isAlive(){
        return $this->getState() == 'running' ;
    }

    public function getHASettings(){
        $ha_status = $this->cluster->getRawData('ha_status','/cluster/ha/status/current');
        foreach($ha_status as $row){
            if( isset($row['sid']) && $row['sid'] == 'vm:'.$this->getID()){
            return json_encode($row);
        }
        }
        return NULL;
    }

    public function isHA(){
        $ha_settings = json_decode($this ->getHASettings(),TRUE);
        return ( $ha_settings['state'] == 'started');
    }

    public function getDiskRead(){
        return $this->getRRDData('diskread');
    }

    public function getDiskWrite(){
        return $this->getRRDData('diskwrite');
    }

    public function getSnapshots(){
        $snapshots = $this->getVMSnapData();
        $pve_keys['name']='name';
        $pve_keys['description']='description';
        $pve_keys['parent']='parent';
        $pve_keys['time']='snaptime';
        $pve_keys['withRAM']='vmstate';

        $ret = array();
        $vmid = $this->getID();
        foreach($snapshots as $row){
                $ret_row = array();
                if( $row['name'] == 'current' ) continue;
                $ret_row['id'] = "$vmid/".$row['name'];
                foreach($pve_keys as $ekey => $pkey ){
                        if( isset($row[$pkey] ) )  $ret_row[$ekey] = $row[$pkey];
                        else $ret_row[$ekey] = NULL;
                }
                $ret_row['withRAM'] = ( isset( $row['withRAM'] ) && $row['withRAM'] == 1 );
                $ret[]=$ret_row;
        }
        return $ret;
    }

    public function getVolumes(){
        $raw_data = $this->getVMRawConfig();
        $storages_list = $this->cluster->getStorages();
        $ret = array();

        foreach($raw_data as $raw_row){
            if( preg_match('/(.*:.*),(size|media)/',$raw_row,$matches)){
                $vol_exp = explode(':',$matches[1],2);
                foreach($storages_list as $storage){
                    if( $storage->getName() == $vol_exp[0]){
                        $volumes = $storage->getVolumes($this->getID());
                        foreach( $volumes as $volume){
                            if($volume['id'] == $matches['1']){
                                $ret[] = $volume;
                                break;
                            }
                        }
                        break;
                    }
                }
            }
        }
        return $ret;
    }

    public function start(){
        $allowed = $this->isAllowedOperation('start');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $state = $this->getState();
        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        if( $state == 'paused' ){
            $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/status/resume");
        }
        else{
            $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/status/start");
        }
    }

    public function shutdown($cleanly=TRUE){
        $allowed = $this->isAllowedOperation('shutdown',$cleanly);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        if( !$cleanly ){
            $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/status/stop");
        }
        else{
            $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/status/shutdown");
        }
    }

    public function reboot($cleanly=TRUE){
        $allowed = $this->isAllowedOperation('reboot',$cleanly);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        if( !$cleanly ){
            $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/status/reset");
        }
        else{
            $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/status/reboot");
        }
    }

    public function migrate($target,$online){
        $allowed = $this->isAllowedOperation('migrate',$online);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        $result = $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/migrate",array('target'=>$target,'online'=>$online));
    }

    public function cloneVM($target_name,$full=TRUE){
        $allowed = $this->isAllowedOperation('clone');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);
        $this->getCluster()->cloneVM($this,$target_name,TRUE);
    }

    public function createSnapshot($name,$description='',$withRam=TRUE){
        $allowed = $this->isAllowedOperation('snapshot',$withRam);
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        $params = ['snapname' => $name,
                        'description' => $description,
                        'vmstate' => $withRam];
        $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/snapshot",$params);
        $this->snap_data = NULL;
    }

    public function reverseToSnapshot($snapid){
        $allowed = $this->isAllowedOperation('snapshot_reverse');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $snapid = explode('/',$snapid);
        $snap_name = $snapid[1];
        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        $this->cluster->getDriver()->post("/nodes/$node/qemu/$vmid/snapshot/$snap_name/rollback");
        $this->snap_data = NULL;
    }

    public function removeSnapshot($snapid){
        $allowed = $this->isAllowedOperation('snapshot_remove');
        if( !$allowed['allowed'] ) throw new Exception($allowed['cause']);

        $snapid = explode('/',$snapid);
        $snap_name = $snapid[1];
        $node = $this->getHost()->getName();
        $vmid = $this->getId();

        $this->cluster->getDriver()->delete("/nodes/$node/qemu/$vmid/snapshot/$snap_name");
        $this->snap_data = NULL;
    }
}

class ProxmoxStorage extends Storage{
    protected $cluster;
    protected $id;

    protected $resource_data=NULL;
    protected $storage_data=NULL;
    protected $volume_data = NULL;

    public function __construct($id,ProxmoxCluster $cluster){
        $this->cluster = $cluster;
        $this->id = $id;
    }

    private function getResourceData(){
        $storage_id = $this->getId();
        if( $this->resource_data == NULL ){
            $resources = $this->cluster->getRawData('cluster_resources_storage','cluster/resources',array('type'=>'storage'));
            foreach($resources as $res){
                if( $res['shared'] == 1 )   $id = 'cluster/'.$res['storage'];
                else                        $id = $res['node'].'/'.$res['storage'];

                if( $id == $storage_id ){
                    if( $this->resource_data == NULL ) {
                        $this->resource_data = $res;
                        $this->resource_data['node'] = array($res['node']);
                    }
                    else{
                        $this->resource_data['node'][] = $res['node'];
                    }
                }
            }
            sort($this->resource_data['node']);
        }
        return $this->resource_data;
    }

    private function getStorageData(){
        if( $this->storage_data == NULL ){
            $storage_id = $this->getId();
            $exp = explode('/',$storage_id);
            $storage = $exp[1];
            $this->storage_data = $this->cluster->getDriver()->get("storage/$storage");
        }
        return $this->storage_data;
    }

    private function getVolumeData(){;
        if( $this->volume_data == NULL ){
            $storage_id = $this->getId();
            $exp = explode('/',$storage_id);
            $storage = $exp[1];

            if( $exp[0] == 'cluster'){
                    $node_id = $this->getHosts()[0]->getName();
            }
            else{
                $node_id = $exp[0];
            }
            $this->volume_data = $this->cluster->getRawData("volumes_$storage_id".'_'.$storage,"nodes/$node_id/storage/$storage/content/");
        }
        return $this->volume_data;
    }

    public function getId(){
        return $this->id;
    }

    public function getName(){
        return $this->getResourceData()['storage'];
    }

    public function getSizeTotal(){
        return $this->getResourceData()['maxdisk'];
    }

    public function getSizeUsed(){
        return $this->getResourceData()['disk'];
    }

    public function getSizeVirtual(){
        $raw_data = $this->getVolumeData();
        $size = 0;
        foreach($raw_data as $row){
            $size += $row['size'];
        }
        return $size;
    }

    /**
      @brief
    */
    public function getContent(){
        return $this->getStorageData()['content'];
    }

    public function getType(){
        return $this->getStorageData()['type'];
    }

    public function getHosts(){
        $ret = array();
        foreach($this->getResourceData()['node'] as $row){
                $ret[] = new ProxmoxHost('node/'.$row,$this->cluster);
        }
        return $ret;
    }

    public function getVolumes($o_vmid='all'){
        $raw_data = $this->getVolumeData();
        $ret = array();
        foreach ($raw_data as $row) {
            if(!isset($row['vmid'])) $row['vmid'] = NULL;
            if( $o_vmid != 'all' && $o_vmid != $row['vmid']) continue;

            $retRow = array();
            $retRow['id'] = $row['volid'];
            if( isset($row['name']) ){
                $retRow['name'] = $row['name'];
            }
            else{
                $retRow['name'] = explode(':',$row['volid'])[1];
            }
            $retRow['desc'] = $row['content'].'('.$row['format'].')';
            $retRow['size_used'] = $row['size'];
            $retRow['size_virt'] = $row['size'];

            $retRow['io_read'] = 0;
            $retRow['io_write'] = 0;

            if( isset($row['parent'])) $retRow['parent'] = $row['parent'];
            else $retRow['parent'] = NULL;

            $retRow['storage'] = $this;
            if( !empty($row['vmid'])) {
                $retRow['vm'] = array($this->cluster->getVM($row['vmid']));
                $vm = $this->cluster->getVM($row['vmid']);
                if( is_object($vm)){
                    $node = $vm->getHost()->getName();
                    $vmid = $row['vmid'];
                    $vm_config = $vm->getVMRawConfig();
                    $n_disks = 0;
                    foreach($vm_config as $key => $value){
                        if( preg_match('/(.*):.*,size=/',$value,$matches) ){
                            $n_disks++;
                        }
                    }
                    $retRow['io_read'] = $vm->getDiskRead()/$n_disks;
                    $retRow['io_write'] = $vm->getDiskWrite()/$n_disks;
                }
            }
            else $retRow['vm'] = array();
            $ret[] = $retRow;
        }
        return $ret;
    }

    public function isShared(){
        return ($this->getResourceData()['shared'] == 1);
    }

    private function getDiskIO($type){
        $volumes = $this->getVolumes();
        $ret = 0;
        foreach($volumes as $volume){
            $ret += $volume[$type];
        }
        return $ret;
    }

    public function getDiskRead(){
        return $this->getDiskIO('io_read');
    }

    public function getDiskWrite(){
        return $this->getDiskIO('io_write');
    }
}
